# tkgeomap_procs.tcl --
#
#	This file provides the tkgeomap_procs package, which augments the
#	tkgeolinearray and tkgeoplace extensions.  See the tkgeomap_procs (n)
#	man page for details.
# 
# Copyright (c) 2004 Gordon D. Carrie. All rights reserved.
# 
# Licensed under the Open Software License version 2.1
# 
# Please address questions and feedback to user0@tkgeomap.org
# 
#  @(#) $Id: tkgeomap_procs.tcl,v 1.36 2006/10/03 19:36:52 tkgeomap Exp $

package require tkgeomap 2
package require tclgeomap_procs 2
package provide tkgeomap_procs 2.10

# geomap::xytolatlon --
#
#	Find latitude and longitude corresponding to canvas point
#	using the map configuration for a canvas geomap_lnarr or
#	geomap_place item.
#
# Arguments:
#	See the tkgeomap_procs (n) man page.
#
# Results:
#	See the tkgeomap_procs (n) man page.

proc geomap::xytolatlon {cvs tagOrId x y} {

    set type [$cvs type $tagOrId]
    if {$type != "geomap_lnarr" && $type != "geomap_place"} {
        error "xytolatlon only works with geomap_lnarr and geomap_place items."
    }
    set coords [$cvs coords $tagOrId]
    set x0 [lindex $coords 0]
    set y0 [lindex $coords 1]
    set refPt [$cvs itemcget $tagOrId -refpoint]
    set proj [$cvs itemcget $tagOrId -projection]
    set rxn [$cvs itemcget $tagOrId -rotation]
    set s [$cvs itemcget $tagOrId -scale]

    # Map coordinates of reference point

    set mpt0 [scalept [rotatpt [$proj fmlatlon $refPt] $rxn] $s]
    set abs0 [lindex $mpt0 0]
    set ord0 [lindex $mpt0 1]

    # Map coordinates of {x y}

    set mPerPx [expr {0.0254 / 72.0 / [tk scaling]}]
    set abs [expr {$abs0 - ($x0 - $x) * $mPerPx}]
    set ord [expr {$ord0 + ($y0 - $y) * $mPerPx}]
    set mpt [scalept [rotatpt [list $abs $ord] \
	    [expr {-1.0 * $rxn}]] [expr {1.0 / $s}]]

    return [$proj tolatlon $mpt]
}

# geomap::latlontoxy --
#
#	Find canvas coordinates corresponding to a {latitude longitude}
#	using the map configuration for a canvas geomap_lnarr or
#	geomap_place item.
#
# Arguments:
#	See the tkgeomap_procs (n) man page.
#
# Results:
#	See the tkgeomap_procs (n) man page.

proc geomap::latlontoxy {cvs tagOrId gpt} {

    set type [$cvs type $tagOrId]
    if {$type != "geomap_lnarr" && $type != "geomap_place"} {
        error "latlontoxy only works with geomap_lnarr and geomap_place items."
    }
    set coords [$cvs coords $tagOrId]
    set x0 [lindex $coords 0]
    set y0 [lindex $coords 1]
    set refPt [$cvs itemcget $tagOrId -refpoint]
    set proj [$cvs itemcget $tagOrId -projection]
    set rxn [$cvs itemcget $tagOrId -rotation]
    set s [$cvs itemcget $tagOrId -scale]

    # Map coordinates of point

    set mpt [scalept [rotatpt [$proj fmlatlon $gpt] $rxn] $s]
    set abs [lindex $mpt 0]
    set ord [lindex $mpt 1]

    # Map coordinates of reference point

    set mpt0 [scalept [rotatpt [$proj fmlatlon $refPt] $rxn] $s]
    set absC [lindex $mpt0 0]
    set ordC [lindex $mpt0 1]

    # Canvas coordinates of gpt

    set pxPerM [expr {72.0 * [tk scaling] / 0.0254}]
    set x [expr {$x0 + ($abs - $absC) * $pxPerM}]
    set y [expr {$y0 - ($ord - $ordC) * $pxPerM}]

    return [list $x $y]
}

# geomap::igeomap --
#
#	This procedure creates an interactive geographic
#	map.  See the user documentation for details.

proc geomap::igeomap {namespaceName pathName args} {
    if { ![regexp {^::} $namespaceName] } {
	# namespaceName is not fully qualified.
	# Concatenate it with name of parent namespace.

	set parent [uplevel 1 {namespace current}]
	if {$parent == "::"} {
	    set namespaceName ::$namespaceName
	} else {
	    set namespaceName ${parent}::$namespaceName
	}
    }
    namespace eval $namespaceName {}
    set ${namespaceName}::frame [frame $pathName]
    namespace eval $namespaceName {

	variable map_ns [namespace current]
	variable canvas [canvas ${frame}.canvas]
	variable toplevel [winfo toplevel $canvas]

	# This array associates the map namespace with its toplevel widget.

	set ::geomap::map_ns($toplevel) $map_ns

	# This bindtag will be used to adjust motion bindings as map projection
	# changes.

	bindtags $canvas [linsert [bindtags $canvas] 0 geomap_motion]

	# Initialize map parameters

	set refpoint {0.0 0.0}
	set projname Mercator
	set projId [::geomap::projection Mercator 0.0]
	set scale 1.0e-7
	set rotation 0.0
	set boundcirclecolor Black
	set update ""
	set bindings {}
	set lazy 0
	set projections {CylEqDist Mercator CylEqArea LambertConfConic \
		LambertEqArea Stereographic PolarStereographic Orthographic}
	set scales {1:10000000 1:20000000 1:30000000 1:45000000 1:60000000 \
		1:90000000 1:120000000}
	set rotationmenu 1

	# Create procedures in the map's namespace.

	# ${map_ns}::xytolatlon --
	#
	#	This map namespace procedure converts x y canvas coordinates in
	#	the map canvas to {latitude longitude}
	#
	# Arguments:
	#	args	- list of canvas coordinates of form
	#		  {?-catch? x1 y1 x2 y2 ...}
	#		  If first word of args is "-catch" skip points that
	#		  cannot be converted.
	#
	# Results:
	#	Return value is a list of form {{lat1 lon1} {lat2 lon2} ...}
	#	giving the geographic coordinates of the canvas points.

	proc ${map_ns}::xytolatlon {args} $::geomap::map_ns_xytolatlon_body

	# ${map_ns}::latlontoxy --
	#
	#	This map namespace procedure converts a lat-lon values to x y
	#	coordinates in the map_canvas window.
	#
	# Arguments:
	#	args	- a list of form {-catch {lat1 lon1} {lat2 lon2} ...}
	#	giving geographic locations.
	#
	# Results:
	#	Return value is a list of form {x1 y1 x2 y2 ...} giving the
	#	canvas coordinates corresponding to the input lat-lon's.

	proc ${map_ns}::latlontoxy {args} $::geomap::map_ns_latlontoxy_body

	# ${map_ns}::setcenter --
	#
	#	This map namespace procedure puts the map refpoint at the canvas
	#	center.
	#
	# Arguments:
	#	None.
	#
	# Results:
	#	coords are set for geomap_lnarr and geomap_place items in the
	#	canvas.
	#	Return value is list of coordinates of map_canvas center.

	proc ${map_ns}::setcenter {} $::geomap::map_ns_setcenter_body

	# map_ns::configure --
	#
	#	This map namespace procedure updates the igeomap associated
	#	with this namespace as indicated by given configuration options.
	#
	# Arguments:
	#	args	- a list of form {-opt arg ?-opt arg ...?} specifying
	#		  the configuration options and values.
	#
	# Results:
	#	Namespace variables and the associated widgets are updated.

	proc ${map_ns}::configure {args} $::geomap::map_ns_configure_body

	# map_ns::cget --
	#
	#	This map namespace procedure retrieves a configuration option.
	#
	# Arguments:
	#	option	- configuration option whose value is sought.
	#
	# Results:
	#	The value of a configuration option.

	proc ${map_ns}::cget {option} $::geomap::map_ns_cget_body

	# ${map_ns}::draw --
	#
	#	This map namespace procedure creates a new geomap_lnarr or
	#	geomap_place item, or re-configures an old one.
	#
	# Arguments:
	#	type	- geomap_lnarr or geomap_place
	#	name	- fully qualified name of the lnarr or place.
	#	args	- list of form {-option arg ?-option arg ...?} giving
	#		  options to pass to the item.  See the tkgeomap (n)
	#		  man page for details.
	#
	# Results:
	#	An item is created or configured.
	#	Return value is the item identifier for new or modified item.

	proc ${map_ns}::draw {type name args} $::geomap::map_ns_draw_body

	# ${map_ns}::erase --
	#
	#	This map namespace procedure deletes and item created with the
	#	draw command.
	#
	# Arguments:
	#	type	- geomap_lnarr or geomap_place
	#	name	- the name of the place or linearray given to the draw
	#		  command.
	#
	# Results:
	#	None.

	proc ${map_ns}::erase {type name} $::geomap::map_ns_erase_body

	# map_ns::set_layers --
	#
	#	This map namespace procedure sets the drawing sequence for
	#	certain canvas items.
	#
	# Arguments:
	#	None.
	#
	# Results:
	#	The drawing sequence for items in the canvas may be changed.

	proc ${map_ns}::set_layers {} $::geomap::map_ns_set_layers_body

	# map_ns::get_layers --
	#
	#	This map namespace procedure retrieves the map's layer list.
	#
	# Arguments:
	#	None.
	#
	# Results:
	#	Return value is the map's layer list.

	proc ${map_ns}::get_layers {} $::geomap::map_ns_get_layers_body

	# map_ns::insert_layer --
	#
	#	This map namespace procedure adds a set of itmes to the
	#	layer list.
	#
	# Arguments:
	#	layer	- name for the layer.  The map layer procedures will
	#		  manage the stacking order for items with tag "$layer".
	#		  If a layer by this name already exists, it is moved.
	#	index	- where in the layer list to insert the new item.
	#		  Interpretation is as for the index argument to
	#		  linsert.
	#
	# Results:
	#	The map's layer list is modified.  The stacking order for items
	#	in the canvas might change.

	proc ${map_ns}::insert_layer {layer {index end}} \
		$::geomap::map_ns_insert_layer_body

	# map_ns::rm_layer --
	# 
	#	This map namespace procedure removes items from the layer list.
	#
	# Arguments:
	#	layer	- the layers to remove.
	#
	# Results:
	#	The map namespace layers list and geomap default layers list
	#	are updated.

	proc ${map_ns}::rm_layer {layer} $::geomap::map_ns_rm_layer_body

	# ::map_ns::layer_cmp --
	#
	#	This procedure sorts layer items in the layer selection
	#	dialog.  It is used as the -command argument to lsort
	#	when putting the items in order.
	#
	# Arguments:
	#	l1, l2	- canvas tags that identify items in the layer dialog.
	#
	# Results:
	#	-1 if item l1 is above item l2 in the window
	#	 0 if l1 and l2 are at the same level (i.e. same item)
	#	 1 if item l1 is below item l2 in the window

	proc ${map_ns}::layer_cmp {l1 l2} $::geomap::map_ns_layer_cmp_body

	# Return cartesian y (distance up from y=0) for an item.

	proc y_cart {cvs item} {
	    set y_cvs [lindex [$cvs coords $item] 1]
	    return [expr {-$y_cvs}]
	}

	# Return canvas y from cartresian y

	proc y_canvas {y_cart} {
	    return [expr {-$y_cart}]
	}

	# Compare two items by distance up.

	proc item_cmp {item1 item2} {
	    variable layer_cvs
	    variable text_id
	    set y1 [y_cart $layer_cvs $text_id($item1)]
	    set y2 [y_cart $layer_cvs $text_id($item2)]
	    expr {$y1 < $y2 ? -1 : $y2 < $y1 ? 1 : 0}
	}

	# ::map_ns::set_layers_dlg --
	#
	#	This procedure creates a window in which the user can
	#	change the layer order by dragging associated canvas items.
	#
	# Arguments:
	#	None.
	#
	# Results:
	#	A dialog appears with items that represent layers.
	#	Map layer sequence changes as user drags items in the dialog.

	proc ${map_ns}::set_layers_dlg {} \
		$::geomap::map_ns_set_layers_dlg_body

	# map_ns::addmenu --
	#
	#	This map namespace procedure adds a menu to the igeomap menu
	#	bar.
	#
	# Arguments:
	#	name	- menu name.  The will be the label on the menu button
	#		  and the part of the path name of the new button and
	#		  menu.
	#		  The name also refers to the menu later, such as when
	#		  it is deleted.
	#		
	# Results:
	#	Return value is the path name of the new menu.
	#	A menu button is added to the menu bar.

	proc ${map_ns}::addmenu {name} $::geomap::map_ns_addmenu_body

	# map_ns::deletemenu --
	#
	#	This map namespace procedure deletes a menu created with addmenu
	#
	# Arguments:
	#	name	- menu name that was given to addmenu
	#
	# Results:
	#	A menu button and its child are destroyed.

	proc ${map_ns}::deletemenu {name} $::geomap::map_ns_deletemenu_body

	# map_ns::ctr_reflon --
	#
	#	This map namespace procedure resets the map when the mouse
	#	button is released.  It is used in bind scripts for certain
	#	projections.
	#
	# Arguments:
	#	None.
	#
	# Results:
	#	Items in the map_canvas canvas are reconfigured.

	proc ${map_ns}::ctr_reflon {} $::geomap::map_ns_ctr_reflon_body

	# map_ns::ctr_refpt --
	#
	#	This map namespace procedure sets the map and projection
	#	reference points to the given point, and resets the bindings.
	#	It is used in bind scripts for reference point projections.
	#
	# Arguments:
	#	x, y	- canvas coordinates of the point which will become
	#		  the new map refpoint and projection reference point.
	#
	# Results:
	#	geomap_lnarr and geomap_place items in the canvas are
	#	reconfigured.

	proc ${map_ns}::ctr_refpt {x y} $::geomap::map_ns_ctr_refpt_body

	# map_ns::set_color --
	#
	#	This map namespace procedure sets a color in the map.
	#
	# Arguments:
	#	elem		- a color element, e.g. land, rivers, places,
	#			  etc. to choose the color for.
	#	value		- what color to use when displaying the element.
	#
	# Results:
	#	Map arrays are updated.  The colorscript for the element
	#	is evaluated.

	proc ${map_ns}::set_color {elem value} $::geomap::map_ns_set_color_body

	# choose_color --
	#
	#	This namespace procedure is the callback for buttons created by
	#	the File->Colors menu.  When activated by a button press, it
	#	prompts the user for a color to use in the map.
	#
	# Arguments:
	#	elem	- a color element, e.g. land, rivers, places,
	#		  etc. to choose the color for.
	#
	# Results:
	#	Return value should be ignored.
	#	colorval($elem) is set to the value chosen by the
	#	user.  The colorval array is dumped to the colorvals list
	#	for later retrieval with the Undo and Redo buttons.

	proc ${map_ns}::choose_color {elem} \
		$::geomap::map_ns_choose_color_body

	# map_ns::color_menu
	#
	#	This map namespace procedure creates a dialog in which the user
	#	can choose new colors for certain map items.
	#
	# Arguments:
	#	None.
	#
	# Results:
	#	A dialog box appears that enables modification of certain canvas
	#	items.  The dialog does not block the application.  It exists
	#	until the user destroys it.

	proc ${map_ns}::color_menu {} $::geomap::map_ns_color_menu_body

	# postscript --
	#
	#	This map namespace procedure generates postscript for the canvas
	#	based on user input from a dialog box.
	#
	# Arguments:
	#	None.
	#
	# Results:
	#	A dialog box appears in which the user gives the name of a
	#	postscript file.  The dialog does not block the application,
	#	and it remains until the user destroys it.

	proc ${map_ns}::postscript {} $::geomap::map_ns_postscript_body

	# ${map_ns}::delete --
	#
	#	This map namespace procedure deletes the map and associated
	#	commands and variables.
	#
	# Arguments:
	#	None.
	#
	# Results:
	#	The namespace for the map, and its associated commands and
	#	variables are deleted.

	# This procedure should be called to delete the map.

	proc ${map_ns}::delete {} $::geomap::map_ns_delete_body

	# These are the canvas coordinates of the map center.

	variable xCtr 0
	variable yCtr 0
	bind $canvas <Configure> [namespace code setcenter]
	setcenter

	# Create a bogus item.  This ensures that calculations that
	# assume the existence of at least one item with the geomap tag
	# will always work.

	draw geomap_place ""

	# Create the menu bar

	variable mbar [frame ${frame}.mbar -relief raised -borderwidth 3]

	# Create the file menu

	menubutton ${frame}.mbar.file -text "File" -menu ${mbar}.file.menu
	menu ${frame}.mbar.file.menu
	${frame}.mbar.file.menu add command -label "Postscript" \
		-command [namespace code postscript]
	${frame}.mbar.file.menu add command -label "Quit" -command {exit}

	# Create the projections menu

	menubutton ${frame}.mbar.projection -text "Projection" \
		-menu ${frame}.mbar.projection.menu
	menu ${frame}.mbar.projection.menu

	# Create the scales menu

	menubutton ${frame}.mbar.scale -text "Scale" -menu ${mbar}.scale.menu
	menu ${frame}.mbar.scale.menu

	# Create the rotation menu

	menubutton ${frame}.mbar.rotation -text "Top" \
		-menu ${frame}.mbar.rotation.menu
	menu ${frame}.mbar.rotation.menu
	foreach r {north ne east se south sw west nw} {
	    ${frame}.mbar.rotation.menu add command -label $r \
		    -command [namespace code [list configure -rotation $r]]
	}
	unset -nocomplain r

	# Pack widgets in main frame.
	# Frame should be positioned by caller.

	pack forget $canvas
	pack ${frame}.mbar.file -side left
	pack ${frame}.mbar -fill x
	pack $canvas -fill both -expand true

	# Apply default configuration

	configure 			\
	    -projname $projname		\
	    -projections $projections	\
	    -scales $scales		\
	    -rotationmenu $rotationmenu
    }

    # Apply user specified configuration

    if {[llength $args] > 0} {
	eval ${namespaceName}::configure $args
    }
}

# geomap::set_motion_bindings --
#
#	This procedure sets the bindings that allow the user to move a map
#	and adjust its reference point.
#
# Arguments:
#	modifier	- event modifier, e.g. "Control" or "Shift" to require
#			  for mouse actions intended to move the map.
#	button		- mouse button to push to move the map.
#
# Results:
#	Bindings to widgets with the geomap_motion tag are modified.

proc geomap::set_motion_bindings {modifier button} {
    set mods [list {} Alt Control Shift]
    if {[lsearch -exact $mods $modifier] < 0} {
	error "Modifier must be one of $mods"
    }
    if {$modifier != ""} {
	set modifier ${modifier}-
    }
    if ![string is integer $button] {
	error "Button identifier must be an integer"
    }

    foreach binding [bind geomap_motion] {
	bind geomap_motion $binding {}
    }

    # Update map bindings.

    set double "<${modifier}Double-Button-$button>"
    set press "<${modifier}Button-$button>"
    set motion "<${modifier}Button$button-Motion>"
    set release "<${modifier}ButtonRelease-$button>"

    # If map projection uses reference longitude, dragging moves the map.

    bind geomap_motion $press [namespace code {
	set tl [winfo toplevel %W]
	set ns $::geomap::map_ns($tl)
	set projname [set ${ns}::projname]
	if {$::geomap::ProjRef($projname) == "longitude"} {
	    namespace eval $ns {
		set x00 %x
		set y00 %y
		set x0 %x
		set y0 %y
	    }
	}
    }]
    bind geomap_motion $motion [namespace code {
	set tl [winfo toplevel %W]
	set ns $::geomap::map_ns($tl)
	set projname [set ${ns}::projname]
	if {$::geomap::ProjRef($projname) == "longitude"} {
	    namespace eval $ns {
		if {[info exists x0] && [info exists y0]} {
		    $canvas move geomap [expr {%x - $x0}] [expr {%y - $y0}]
		}
		set x0 %x
		set y0 %y
	    }
	}
    }]
    bind geomap_motion $release [namespace code {
	set tl [winfo toplevel %W]
	set ns $::geomap::map_ns($tl)
	set projname [set ${ns}::projname]
	if {$::geomap::ProjRef($projname) == "longitude"} {
	    namespace eval $ns {
		if {[info exists x00] && [info exists y00]} {
		    if {%x != $x00 && %y != $y00} {
			ctr_reflon
		    }
		    unset -nocomplain x00
		    unset -nocomplain y00
		}
		unset -nocomplain x0
		unset -nocomplain y0
	    }
	}
    }]

    # If map projection uses a reference point, double clicking moves
    # the map.

    bind geomap_motion $double [namespace code {
	set tl [winfo toplevel %W]
	set ns $::geomap::map_ns($tl)
	set projname [set ${ns}::projname]
	if {$::geomap::ProjRef($projname) == "point"} {
	    namespace eval $ns {
		ctr_refpt %x %y
	    }
	}
    }]
}

# geomap::get_motion_bindings --
#
#	This procedure sets the bindings that allow the user to move a map
#	and adjust its reference point.
#
# Arguments:
#	None.
#
# Results:
#	Return value is a list of bindings for the geomap_motion tag.

proc geomap::get_motion_bindings {} {
    return [bind geomap_motion]
}

# This script is the body for the mapNamespace::xytolatlon procedures.
# See the procedure definition above for documentation.
# Signature: xytolatlon {args}

set geomap::map_ns_xytolatlon_body {
    variable canvas
    variable refpoint
    variable projId
    variable rotation
    variable scale

    if {[lindex $args 0] == "-catch"} {
	set catch 1
	set args [lrange $args 1 end]
    } else {
	set catch 0
    }

    # Map coordinates of reference point

    set mpt0 [::geomap::scalept [::geomap::rotatpt \
	    [$projId fmlatlon $refpoint] $rotation] $scale]
    set abs0 [lindex $mpt0 0]
    set ord0 [lindex $mpt0 1]

    # Map coordinates of args = {x y ...}

    set coords [$canvas coords geomap]
    set x0 [lindex $coords 0]
    set y0 [lindex $coords 1]
    set mPerPx [expr {0.0254 / 72.0 / [tk scaling]}]
    set l [llength $args]
    if {$l == 2} {
	set x [lindex $args 0]
	set y [lindex $args 1]
	set abs [expr {$abs0 - ($x0 - $x) * $mPerPx}]
	set ord [expr {$ord0 + ($y0 - $y) * $mPerPx}]
	set mpt [::geomap::scalept [::geomap::rotatpt [list $abs $ord] \
		[expr {-1.0 * $rotation}]] [expr {1.0 / $scale}]]
	if $catch {
	    if {[catch [list $projId tolatlon $mpt] gpt] == 0} {
		return $gpt
	    } else {
		return {}
	    }
	} else {
	    return [$projId tolatlon $mpt]
	}
    } elseif {$l % 2 == 0} {
	set r {}
	if $catch {
	    foreach {x y} $args {
		set abs [expr {$abs0 - ($x0 - $x) * $mPerPx}]
		set ord [expr {$ord0 + ($y0 - $y) * $mPerPx}]
		set mpt [::geomap::scalept [::geomap::rotatpt [list $abs $ord] \
			[expr {-1.0 * $rotation}]] [expr {1.0 / $scale}]]
		if {[catch [list $projId tolatlon $mpt] gpt] == 0} {
		    lappend r $gpt
		}
	    }
	} else {
	    foreach {x y} $args {
		set abs [expr {$abs0 - ($x0 - $x) * $mPerPx}]
		set ord [expr {$ord0 + ($y0 - $y) * $mPerPx}]
		set mpt [::geomap::scalept [::geomap::rotatpt [list $abs $ord] \
			[expr {-1.0 * $rotation}]] [expr {1.0 / $scale}]]
		lappend r [$projId tolatlon $mpt]
	    }
	}
	return $r
    } else {
	error "Number of coordinates cannot be odd"
    }
}

# This script is the body for the mapNamespace::latlontoxy procedures.
# See the procedure definition above for documentation.
# Signature: latlontoxy {args}

set geomap::map_ns_latlontoxy_body {
    variable canvas
    variable refpoint
    variable projId
    variable rotation
    variable scale

    if {[lindex $args 0] == "-catch"} {
	set catch 1
	set args [lrange $args 1 end]
    } else {
	set catch 0
    }

    set coords [$canvas coords geomap]
    set x0 [lindex $coords 0]
    set y0 [lindex $coords 1]
    set pxPerM [expr {72.0 * [tk scaling] / 0.0254}]

    # Map coordinates of reference point

    set mpt0 [::geomap::scalept [::geomap::rotatpt \
	    [$projId fmlatlon $refpoint] $rotation] $scale]
    set absC [lindex $mpt0 0]
    set ordC [lindex $mpt0 1]

    # Map coordinates of point

    set r {}
    if $catch {
	foreach geoPt $args {
	    if {[catch [::geomap::scalept [::geomap::rotatpt \
		    [$projId fmlatlon $geoPt] $rotation] $scale] mpt] == 0} {
		set abs [lindex $mpt 0]
		set ord [lindex $mpt 1]
		set x [expr {$x0 + ($abs - $absC) * $pxPerM}]
		set y [expr {$y0 - ($ord - $ordC) * $pxPerM}]
		lappend r $x $y
	    }
	}
    } else {
	foreach geoPt $args {
	    set mpt [::geomap::scalept [::geomap::rotatpt \
		    [$projId fmlatlon $geoPt] $rotation] $scale]
	    set abs [lindex $mpt 0]
	    set ord [lindex $mpt 1]
	    set x [expr {$x0 + ($abs - $absC) * $pxPerM}]
	    set y [expr {$y0 - ($ord - $ordC) * $pxPerM}]
	    lappend r $x $y
	}
    }
    return $r
}

# This script is the body for the mapNamespace::setcenter procedures.
# See the procedure definition above for documentation.
# Signature: setcenter {}

set geomap::map_ns_setcenter_body {
    variable canvas
    variable xCtr
    variable yCtr

    set xCtr [expr {[winfo width $canvas] / 2}]
    set yCtr [expr {[winfo height $canvas] / 2}]
    foreach item [$canvas find withtag geomap] {
	$canvas coords $item $xCtr $yCtr
    }
    return [list $xCtr $yCtr]
}

# This script is the body for map_ns::configure procedures.
# See the procedure definition above for documentation.
# Signature: configure {args}

set geomap::map_ns_configure_body {
    variable frame
    variable canvas
    variable mbar
    variable refpoint
    variable projname
    variable projId
    variable scale
    variable rotation
    variable lazy
    variable projections
    variable scales
    variable rotationmenu
    variable bindings
    variable update
    variable boundcirclecolor

    foreach {opt arg} $args {
	switch -- $opt {
	    -refpoint {
		$canvas itemconfigure geomap -refpoint $arg
		set refpoint $arg
		uplevel #0 $update
	    }
	    -projname {
		if {[lsearch $::geomap::ProjNames $arg] < 0} {
		    error "Projection must be one of $::geomap::ProjNames"
		}
		set projname $arg

		# Delete old bound circle 

		set bound_circle [namespace current]::bound_circle
		if {[namespace which -command $bound_circle] != ""} {
		    rename $bound_circle {}
		    erase geomap_lnarr $bound_circle
		}

		# Set projection

		if {$::geomap::ProjRef($projname) == "longitude"} {
		    set lon [::geomap::longitude $refpoint]
		    eval $projId set $projname $lon
		} elseif {$::geomap::ProjRef($projname) == "point"} {
		    if {$projname == "PolarStereographic"} {
			set lat [::geomap::latitude $refpoint]
			if {$lat > 0.0} {
			    set refpoint {90.0 0.0}
			    eval $projId set PolarStereographic N
			} else {
			    set refpoint {-90.0 0.0}
			    eval $projId set PolarStereographic S
			}
		    } else {
			eval $projId set $projname $refpoint
		    }
		} else {
		    error "Unable to set $projname projection"
		}

		# If necessary, draw new bound circle

		if {$::geomap::ProjRef($projname) == "point"
			&& $::geomap::ProjDomain($projname) == "hemisphere"} {
		    ::geomap::lnarr fmlist $bound_circle \
			    [::geomap::circle $refpoint 89.9]
		    draw geomap_lnarr $bound_circle -outline $boundcirclecolor \
			    -tags "bound_circle"
		}

		# Run update script

		uplevel #0 $update
	    }
	    -scale {
		if ![string is double $arg] {
		    set scale [::geomap::cartg $arg]
		} else {
		    set scale $arg
		}
		$canvas itemconfigure geomap -scale $scale
		uplevel #0 $update
	    }
	    -rotation {
		$canvas itemconfigure geomap -rotation $arg
		set rotation [$canvas itemcget geomap -rotation]
		uplevel #0 $update
	    }
	    -boundcirclecolor {
		$canvas itemconfigure bound_circle -outline $arg
		set boundcirclecolor $arg
	    }
	    -update {
		set update $arg
	    }
	    -colormenu {
		if $arg {
		    if {[catch [list ${mbar}.file.menu index "Colors"]] != 0} {
			${mbar}.file.menu insert 0 command -label "Colors" \
				-command [namespace code color_menu] 
		    }
		}
	    }
	    -layermenu {
		if $arg {
		    if {[catch [list ${mbar}.file.menu index "Layers"]] != 0} {
			${mbar}.file.menu insert 0 command -label "Layers" \
				-command [namespace code set_layers_dlg] 
		    }
		}
	    }
	    -lazy {
		if ![string is boolean $arg] {
		    error "lazy must be boolean"
		}
		set lazy $arg
	    }
	    -projections {
		# Replace the selections in the Projections menu
		# with the given list.

		set projections $arg
		${frame}.mbar.projection.menu delete 2 end
		if {[llength $projections] == 0} {
		    pack forget ${frame}.mbar.projection
		} else {
		    foreach p $projections {
			${frame}.mbar.projection.menu add command \
				-label $p -command [namespace code \
					[list configure -projname $p]]
		    }
		    pack ${frame}.mbar.projection -side left
		}
	    }
	    -scales {
		# Replace the selections in the Scales menu
		# with the given list.

		set scales $arg
		${frame}.mbar.scale.menu delete 0 end
		if {[llength $scales] == 0} {
		    pack forget ${frame}.mbar.scale
		} else {
		    foreach s $scales {
			${frame}.mbar.scale.menu add command \
				-label $s -command [namespace code \
					[list configure -scale $s]]
		    }
		    pack ${frame}.mbar.scale -side left
		}
	    }
	    -rotationmenu {
		# Present or hide the rotations menu as requested.

		set rotationmenu $arg
		if {$rotationmenu} {
		    pack ${frame}.mbar.rotation -side left
		} else {
		    pack forget ${frame}.mbar.rotation
		}
	    }
	    default {
		if {[catch [list $canvas configure $opt $arg] msg] != 0} {
		    error "Could not configure $opt"
		}
	    }
	}
    }
}

# This script is the body for map_ns::cget procedures.
# See the procedure definition above for documentation.
# Signature: cget {option}

set geomap::map_ns_cget_body {
    variable canvas
    variable refpoint
    variable projname
    variable projId
    variable scale
    variable rotation
    variable boundcirclecolor
    variable lazy
    variable projections
    variable scales
    variable rotationmenu

    switch -- $option {
	-refpoint {
	    return $refpoint
	}
	-projname {
	    return $projname
	}
	-scale {
	    return $scale
	}
	-rotation {
	    return $rotation
	}
	-boundcirclecolor {
	    return $boundcirclecolor
	}
	-projname {
	    return $projname
	}
	-lazy {
	    return $lazy
	}
	-projections {
	    return $projections
	}
	-scales {
	    return $scales
	}
	-rotationmenu {
	    return $rotationmenu
	}
	default {
	    if {[catch [list $canvas cget $opt $arg] msg] != 0} {
		error "Unknown configuration option $opt"
	    }
	}
    }
}

# This script is the body for the mapNamespace::draw procedures.
# See the procedure definition above for documentation.
# Signature: draw {type name args}

set geomap::map_ns_draw_body {
    variable canvas
    variable xCtr
    variable yCtr
    variable refpoint
    variable projId
    variable scale
    variable rotation

    array set config $args
    switch -exact -- $type {
	geomap_lnarr {
	    array set config [list -lnarr $name -refpoint $refpoint	\
		    -projection $projId	-scale $scale -rotation $rotation]
	    if {$name != ""
		    && [set id [$canvas find withtag $name]] != ""
		    && [$canvas type $id] == "geomap_lnarr"} {
		# Configure pre-existing geomap_lnarr item.

		eval $canvas itemconfigure $id [array get config]
	    } else {
		# Create new geomap_lnarr item.

		set id [eval $canvas create geomap_lnarr \
			$xCtr $yCtr [array get config]]
	    }
	}
	geomap_place {
	    array set config [list -place $name -refpoint $refpoint	\
		    -projection $projId	-scale $scale -rotation $rotation]
	    if {$name != ""
		    && [set id [$canvas find withtag $name]] != ""
		    && [$canvas type $id] == "geomap_place"} {
		# Configure pre-existing geomap_place item.

		eval $canvas itemconfigure $id $args
	    } else {
		# Create new geomap_place item.

		set id [eval $canvas create geomap_place \
			$xCtr $yCtr [array get config]]
	    }
	}
	default {
	    error "Type must be \"geomap_lnarr\" or \"geomap_place\""
	}
    }
    $canvas addtag $name withtag $id
    $canvas addtag geomap withtag $id
    return $id
}

# This script is the body for the mapNamespace::erase procedures.
# See the procedure definition above for documentation.
# Signature: erase {type name}

set geomap::map_ns_erase_body {
    variable canvas

    if {$type != "geomap_lnarr" && $type != "geomap_place"} {
	error "Type must be \"geomap_lnarr\" or \"geomap_place\""
    }
    foreach id [$canvas find withtag $name] {
	if {[$canvas type $id] == $type} {
	    $canvas delete $id
	    break
	}
    }
}

# This script is the body for the mapNamespace::set_layers procedures.
# See the procedure definition above for documentation.
# Signature: set_layers {}

set geomap::map_ns_set_layers_body {
    variable canvas
    variable layers

    if { ![info exists layers] } {
	return
    }
    foreach layer $layers {
	if {[llength [$canvas find withtag $layer]] > 0} {
	    if { [info exists aboveThis] } {
		$canvas raise $layer $aboveThis
	    } else {
		$canvas raise $layer
	    }
	    set aboveThis $layer
	}
    }


    # Update layer dialog if present

    variable layer_dlg
    if {[info exists layer_dlg] && ![winfo exists $layer_dlg]} {
	set_layers_dlg
    }
}

# This script is the body for the mapNamespace::get_layers procedures.
# See the procedure definition above for documentation.
# Signature: get_layers {}

set geomap::map_ns_get_layers_body {
    variable layers

    if { ![info exists layers] } {
	set layers {}
    }
    return $layers
}

# This script is the body for the mapNamespace::insert_layer procedures.
# See the procedure definition above for documentation.
# Signature: insert_layer {layer {index end}

set geomap::map_ns_insert_layer_body {
    variable layers

    if { ![info exists layers] } {
	set layers {}
    }
    if {[lsearch $layers $layer] >= 0} {
	rm_layer $layer
    }
    set layers [linsert $layers $index $layer]
    set_layers
}

# This script is the body for the mapNamespace::rm_layer procedures.
# See the procedure definition above for documentation.
# Signature: rm_layer {layer}

set geomap::map_ns_rm_layer_body {
    variable layers

    if { ![info exists layers] } {
	return
    }
    set l [lsearch $layers $layer]
    set layers [lreplace $layers $l $l]
}

# This script is the body for the mapNamespace::set_layers_dlg procedures.
# See the procedure definition above for documentation.
# Signature: set_layers_dlg {}

set geomap::map_ns_set_layers_dlg_body {
    variable margin 24
    variable layers
    variable layer_dlg
    variable layer_cvs
    variable map
    variable margin
    variable items
    variable item_fm_id
    variable text_id
    variable box_id
    variable line_id
    variable all_children
    variable children
    variable leaves
    variable column
    variable layer_dlg .${map}_layers
    variable layer_cvs ${layer_dlg}.c

    # Create the toplevel widget and canvas.

    if { ![winfo exists $layer_dlg] } {
	toplevel $layer_dlg
    } else {
	raise $layer_dlg
    }
    if { ![winfo exists $layer_cvs] } {
	canvas $layer_cvs
    }

    # Create or recreate items

    $layer_cvs delete layer

    set widest 0.0
    set tallest 0.0
    set c_max -1

    set items {}
    array unset column
    foreach layer $layers {
	set parts [split $layer :]
	set item ""
	foreach part $parts {
	    if {$item == ""} {
		set item $part
	    } else {
		append item :$part
	    }
	    set l_item [split $item :]
	    set c [llength $l_item]
	    set c_max [expr {($c > $c_max) ? $c : $c_max}]
	    if { ![info exists column($c)] || [lsearch $column($c) $item] < 0} {
		lappend column($c) $item
	    }
	    if {[lsearch $items $item] < 0} {
		set txt [lindex $l_item end]
		set id [$layer_cvs create text 0 0 -text $txt \
			-fill yellow -anchor w -tags [list layer text $item]]
		set text_id($item) $id
		set item_fm_id($id) $item
		set font_desc [$layer_cvs itemcget $id -font]
		set asc [font metrics $font_desc -ascent]
		set des [font metrics $font_desc -descent]
		set h [expr {$asc + $des + 4}]
		set tallest [expr {($h > $tallest) ? $h : $tallest}]
		set w [font measure $font_desc $txt]
		set widest [expr {($w > $widest) ? $w : $widest}]
		lappend items $item
	    }
	}
    }
    foreach item $items {
	    set children($item) \
		    [lsearch -all -inline -regexp $items "^${item}:\[^:\]+$"]
	    set all_children($item) [lsearch -all -inline $items ${item}:*]
	    set leaves($item) [lsearch -all -inline -regexp $layers "^${item}"]
    }

    # Postion items

    set y $margin
    foreach layer $layers {
	set c [llength [split $layer :]]
	set x [expr {$margin + ($c - 1) * $widest + 3}]
	$layer_cvs coords $text_id($layer) $x [y_canvas $y]
	set y [expr {$y + $tallest + 3}]
    }

    set jj [array names column]
    set l [llength $jj]
    foreach j [lrange [lsort -integer -decreasing $jj] 1 [expr {$l - 1}]] {
	foreach item $column($j) {
	    if {[llength $children($item)] > 0} {
		# Position parent halfway up column of children

		set c [llength [split $item :]]
		set x [expr {$margin + ($c - 1) * $widest + 3}]
		set ids {}
		foreach child $children($item) {
		    lappend ids $text_id($child)
		}
		set bbox [eval $layer_cvs bbox $ids]
		set y1 [lindex $bbox 1]
		set y2 [lindex $bbox 3]
		set y [expr {0.5 * ($y1 + $y2) + 3}]
		$layer_cvs coords $text_id($item) $x $y

		# Draw lines from parent to children

		set p1 [$layer_cvs coords $text_id($item)]
		foreach child $children($item) {
		    set p2 [$layer_cvs coords $text_id($child)]
		    set id [$layer_cvs create line [concat $p1 $p2] \
			    -fill black -tags [list layer line $item]]
		    set parent_id $text_id($item)
		    set child_id $text_id($child)
		    set line_id(${parent_id},${child_id}) $id
		}
	    }
	}
    }

    # Give the text a black background

    foreach item $items {
	set bbox [$layer_cvs bbox $text_id($item)]
	set id [$layer_cvs create rectangle $bbox -fill black \
		-tags [list layer rect $item]]
	set box_id($item) $id
	set item_fm_id($id) $item
    }
    $layer_cvs raise layer&&text

    # Make everything visible.

    set bbox [$layer_cvs bbox layer]
    set y1 [expr {[lindex $bbox 1]}]
    $layer_cvs move layer 0 [expr {-$y1 + $margin}]
    set bbox [$layer_cvs bbox layer]
    set x2 [expr {[lindex $bbox 2] + $margin}]
    set y2 [expr {[lindex $bbox 3] + $margin}]
    $layer_cvs configure -width $x2 -height $y2
    pack $layer_cvs

    # The items representing the layers can be moved up
    # and down by dragging.  At button release, the new
    # layer sequence is applied to the map.

    $layer_cvs bind layer&&(text||rect) <Button-1> [namespace code {
	set y0 %y
    }]
    $layer_cvs bind layer&&(text||rect) <B1-Motion> [namespace code {
	set dy [expr {%y - $y0}]
	set curr_id [%W find withtag current]
	set item $item_fm_id($curr_id)
	$layer_cvs move $text_id($item) 0 $dy
	$layer_cvs move $box_id($item) 0 $dy
	foreach child $all_children($item) {
	    $layer_cvs move $text_id($child) 0 $dy
	    $layer_cvs move $box_id($child) 0 $dy
	}

	# Update lines

	foreach item $items {
	    if {[llength $children($item)] > 0} {
		set p1 [$layer_cvs coords $text_id($item)]
		foreach child $children($item) {
		    set p2 [$layer_cvs coords $text_id($child)]
		    set parent_id $text_id($item)
		    set child_id $text_id($child)
		    set id $line_id(${parent_id},${child_id})
		    $layer_cvs coords $id [concat $p1 $p2]
		}
	    }
	}

	set y0 %y
    }]
    $layer_cvs bind layer&&(text||rect) <ButtonRelease-1> [namespace code {
	# At end of drag, rearrange layer order according to the vertical
	# sequence in the layer display.

	# siblings are items at the same depth and with the same parent
	# as the item that just moved.  Only the dragged item, its siblings,
	# and their associated children will move, and they will only move
	# relative to each other.  Thus, an item cannot be dragged to another
	# branch in the layer hierarchy.

	# The all_leaves list contains layers descended from the item and its
	# siblings.  It is a sublist from the layers list.  all_leaves is
	# removed from the layers list, re-ordered according to the vertical
	# sequence of siblings in the display (determined by the item_cmp proc),
	# and reinserted into the layers list at the same place.

	set curr_id [%W find withtag current]
	set item $item_fm_id($curr_id)
	if [regexp {(.*):} $item m parent] {
	    set siblings $children($parent)
	} else {
	    set siblings $column(1)
	}
	set all_leaves {}
	foreach sibling $siblings {
	    eval lappend all_leaves $leaves($sibling)
	}
	set start [lsearch $layers [lindex $all_leaves 0]]
	set len [llength $all_leaves]
	set last [expr {$start + $len - 1}]
	set layers [lreplace $layers $start $last]
	set siblings [lsort -command item_cmp $siblings]
	set all_leaves {}
	foreach sibling $siblings {
	    eval lappend all_leaves $leaves($sibling)
	}
	set layers [eval linsert \$layers $start $all_leaves]
	set_layers
    }]
}

# This script is the body for the mapNamespace::layer_cmp procedures.
# See the procedure definition above for documentation.
# Signature: layer_cmp {l1 l2}

set geomap::map_ns_layer_cmp_body {
    variable layer_cvs
    set y1 [lindex [$layer_cvs coords $l1] 1]
    set y2 [lindex [$layer_cvs coords $l2] 1]
    expr {$y1 < $y2 ? -1 : $y2 < $y1 ? 1 : 0}
}

# This script is the body for map_ns::addmenu procedures.
# See the procedure definition above for documentation.
# Signature: addmenu {name}

set geomap::map_ns_addmenu_body {
    variable frame

    if [regexp {[ \f\t\n\r]} $name] {
	error "Menu name cannot contain whitespace"
    }
    set lName [string tolower $name]
    set button_path ${frame}.mbar.$lName
    set menu_path ${button_path}.menu
    if { [winfo exists $button_path] } {
	return $menu_path
    }
    menubutton $button_path -text $name -menu $menu_path
    menu $menu_path
    pack $button_path -side left
    return $menu_path
}

# This script is the body for map_ns::deletemenu procedures.
# See the procedure definition above for documentation.
# Signature: deletemenu {name}

set geomap::map_ns_deletemenu_body {
    variable frame

    destroy ${frame}.mbar.[string tolower $name]
}

# This script is the body for map_ns::ctr_reflon procedures.
# See the procedure definition above for documentation.
# Signature: ctr_reflon {}

set geomap::map_ns_ctr_reflon_body {
    variable canvas
    variable projname
    variable lazy

    set toplevel [winfo toplevel $canvas]
    set oldCursor [$toplevel cget -cursor]
    $toplevel configure -cursor watch
    update

    set xCtr [expr {[winfo width $canvas] / 2}]
    set yCtr [expr {[winfo height $canvas] / 2}]
    configure -refpoint [xytolatlon $xCtr $yCtr]
    foreach item [$canvas find withtag geomap] {
	$canvas coords $item $xCtr $yCtr
    }
    if !$lazy {
	configure -projname $projname
    }

    $toplevel configure -cursor $oldCursor
}

# This script is the body for map_ns::ctr_refpt procedures.
# See the procedure definition above for documentation.
# Signature: ctr_refpt {x y}

set geomap::map_ns_ctr_refpt_body {
    variable canvas
    variable projname

    set toplevel [winfo toplevel $canvas]
    set oldCursor [$toplevel cget -cursor]
    $toplevel configure -cursor watch
    update

    configure -refpoint [xytolatlon $x $y]
    if {$projname == "PolarStereographic"} {
	configure -projname Stereographic
    } else {
	configure -projname $projname
    }

    $toplevel configure -cursor $oldCursor
}

# This script is the body for map_ns::set_color procedures.
# See the procedure definition above for documentation.
# Signature: set_color {elem value}

set geomap::map_ns_set_color_body {
    variable colorval
    variable colorscript
    variable colorvals

    # Save the current color configuration in the colorvals list

    if { [info exists colorvals] } {
	set colorvals [linsert $colorvals 0 [array get colorval]]
    } else {
	set colorvals [list [array get colorval]]
    }

    # Update the colored element with user selection

    set colorval($elem) $value
    namespace eval :: $colorscript($elem)
}

# This script is the body for map_ns::choose_color procedures.
# See the procedure definition above for documentation.
# Signature: choose_color {elem}

set geomap::map_ns_choose_color_body {
    variable colorval
    variable colorscript
    variable colorvals
    variable color_dlg

    set map [namespace tail [namespace current]]
    set new [tk_chooseColor -title "$map $elem" \
	    -initialcolor $colorval($elem)]
    if {[string length $new] > 0} {
	set_color $elem $new
	set btn ${color_dlg}.[string tolower $elem]
	$btn configure -background $colorval($elem)
	${color_dlg}.f.undo configure -state normal
    }
}

# This script is the body for map_ns::color_menu procedures.
# See the procedure definition above for documentation.
# Signature: color_menu {}

set geomap::map_ns_color_menu_body {
    variable colorval
    variable colorscript
    variable colorvals

    set map [namespace tail [namespace current]]
    variable color_dlg .${map}_colors
    if { [winfo exists $color_dlg] } {
	# Dialog already exists.  No further action is needed.

	raise $color_dlg
	return
    }

    # Create a dialog box

    toplevel $color_dlg

    # For each color element in the map, create a
    # button to modify the color.

    foreach elem [lsort [array names colorval]] {
	set btn $color_dlg.[string tolower $elem]
	button $btn -text $elem -background $colorval($elem) \
		-command [namespace code [list choose_color $elem]]
	pack $btn -fill x
    }

    # Button row for action buttons

    frame ${color_dlg}.f
    button ${color_dlg}.f.load -text "Load" -command [namespace code {
	set f [tk_getOpenFile]

	# Search the input file for a line starting with "colors:"
	# and set the colorval array from its contents.

	if {$f != ""} {
	    set in [open $f]
	    while {[gets $in line] >= 0} {
		if [regexp {^colors:[ 	]*(.*)} $line m arr] {

		    # Save the current color configuration in the
		    # colorvals list

		    if { [info exists colorvals] } {
			set colorvals \
				[linsert $colorvals 0 [array get colorval]]
		    } else {
			set colorvals [list [array get colorval]]
		    }
		    ${color_dlg}.f.undo configure -state normal

		    # Set the new color configuration

		    array set colorval $arr
		    foreach elem [array names colorval] {
			set btn ${color_dlg}.[string tolower $elem]
			$btn configure -background $colorval($elem)
			namespace eval :: $colorscript($elem)
		    }
		    unset -nocomplain btn
		}
	    }
	    close $in
	    unset in
	}
	unset f
    }]
    button ${color_dlg}.f.save -text "Save" -command [namespace code {
	set f [tk_getSaveFile]
	if {$f != ""} {
	    set out [open $f w]
	    puts $out "
# Map colors --
# The list following \"colors:\" below specifies map color scheme.
# The list is a set of key value pairs.  The keys identify items to be
# colored, such as \"water\" or \"land.\"  Each key is followed by a color
# value, such as \"Blue4\" or \"wheat\" indicating the color to use for the
# corresponding item.
	    "
	    puts $out "colors: [array get colorval]"
	    close $out
	    unset out
	}
	unset f
    }]

    # "Undo" button moves one element down in the colorvals list and
    # displays the color scheme stored in that element in the dialog.
    # If we end up at start of colorvals list, dim the "Undo" button.

    if {[info exists colorvals] && [llength $colorvals] > 0} {
	set st "normal"
    } else {
	set st "disabled"
    }
    button ${color_dlg}.f.undo -text "Undo" -state $st \
	    -command [namespace code {
		array set colorval [lindex $colorvals 0]
		set colorvals [lrange $colorvals 1 end]
		if {[llength $colorvals] == 0} {
		    ${color_dlg}.f.undo configure -state disabled
		}
		foreach elem [array names colorval] {
		    set btn ${color_dlg}.[string tolower $elem]
		    $btn configure -background $colorval($elem)
		    if { [info exists colorscript($elem)] } {
			namespace eval :: $colorscript($elem)
		    }
		}
	    }]

    pack ${color_dlg}.f.undo ${color_dlg}.f.save ${color_dlg}.f.load -side left
    pack ${color_dlg}.f
}

# This script is the body for map_ns::postscript procedures.
# See the procedure definition above for documentation.
# Signature: postscript {}

set geomap::map_ns_postscript_body {
    variable canvas
    variable postscript_file
    variable color_mode

    if { [winfo exists .ps] } {
	# Dialog already exists.  No further action is needed.

	raise .ps
	return
    }

    # Create a dialog box prompting user for a name for a postscript file.

    toplevel .ps
    set x [expr {[winfo x $canvas] + 200}]
    set y [expr {[winfo y $canvas] + 200}]
    wm geometry .ps +$x+$y
    label .ps.msg -text "Send postscript to file"
    entry .ps.e
    if ![info exists postscript_file] {
	set postscript_file "c.ps"
    }
    .ps.e insert 0 $postscript_file
  
    # Radio buttons for colormode

    if { ![info exists color_mode] } {
	set color_mode gray
    }
    frame .ps.r
    set ns [namespace current]
    radiobutton .ps.r.color -text color -variable ${ns}::color_mode -value color
    radiobutton .ps.r.gray -text gray -variable ${ns}::color_mode -value gray
    radiobutton .ps.r.mono -text mono -variable ${ns}::color_mode -value mono
    pack .ps.r.color .ps.r.gray .ps.r.mono -side left -expand 1 -fill x
  
    # OK and Cancel

    frame .ps.b
    bind .ps.e <Return> [namespace code {
	    set postscript_file [.ps.e get]
	    if {[string length $postscript_file] == 0} {
		tk_messageBox -title "Error" -type ok \
			-message "Must have postscript file name"
		destroy .ps
		return
	    }
	    $canvas postscript -colormode $color_mode -file $postscript_file
	    destroy .ps
	}]
    button .ps.b.ok -text OK -command [bind .ps.e <Return>]
    button .ps.b.cancel -text Cancel -command {destroy .ps}

    pack .ps.b.ok .ps.b.cancel -side left
    pack .ps.msg .ps.e .ps.r .ps.b
}

# This script is the body for map_ns::delete procedures.
# See the procedure definition above for documentation.
# Signature: delete {}

set geomap::map_ns_delete_body {
    variable canvas
    variable projId

    destroy $canvas
    rename $projId {}
    namespace delete [namespace current]
}
